 // This unit is used, in order to manipulate the JFL file content

unit JFLFile;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Dialogs, StdCtrls, ComCtrls, Unit2;

type
  TJFLFileEntries = array of RECORD
                      Offset: LongInt;
                      Hlam0: LongInt;
                      Size: LongInt;
                      Hlam1: LongInt;
                      Name: ARRAY[1..200]OF Char;
                    END;
  TJFLFileName = ARRAY[1..200]OF Char;
  TJFLBlockContent = ARRAY[1..128]OF Byte;
  TOpenCharArray = array of Char;

  TJFRLEsearch = RECORD
                      repeating : byte;
                      position : byte;
                      number : byte;
                 end;

   TJFLSegmentContent = Record
                         segmenttype : byte;
                         segmentlen : byte;
                         segment: TJFLBlockContent;
                 end;

var
   JFLfileStream : TFileStream;      // The stream towards the JFL file
   JFLFileEntries : TJFLFileEntries; // All file indexing entries
   JFLFileName : String;             // The Filename of the JFL file
   TotalFiles: LongInt;              // Total number of files contained
   
   memo: TMemo;                      // The memo panel for infos
   progress: TProgressBar;           // File progress bar
   progress2: TProgressBar;           // File progress bar

   // prepare everything regarding JFL file
   procedure InitJFLForReading(fileName : string);
   procedure InitJFLForWriting(fileName : string);
   procedure UninitJFL();  // release JFL file
   
   procedure createFullJFLFromDirectory(dirpath: string);
   procedure extractFullJFLtoDirectory(dirpath: string);
   
   function checkJFLstructure() :  TStringList;  // header check
   
   procedure initJFLMemoBox(m: TMemo); // Memo link
   procedure initJFLFileProgress(p: TProgressBar); // File progress link
   procedure initJFLAdditionalProgress(p: TProgressBar); // Additional progress link

   // procedure RLECompress(var block : TJFLBlockContent;var olen : byte); // test

implementation

/////////////////////////////////////
//                                 //
//    Visual indicators - begin    //
//                                 //
/////////////////////////////////////

procedure initJFLMemoBox(m: TMemo);
begin
     memo := m;
end;

procedure initJFLFileProgress(p: TProgressBar);
begin
     progress := p;
end;

procedure initJFLAdditionalProgress(p: TProgressBar);
begin
     progress2 := p;
end;

/////////////////////////////////////
//                                 //
//    Visual indicators - end      //
//                                 //
/////////////////////////////////////

// Converts the filename to string, and cleans rubbish
function convertJFLFileNameToString(c: TJFLFileName) : string;
var i : integer;
    s : string;
begin
     s:= Copy(c,1,Pos(#$00,c)-1);
     result := s;
end;

// Get the first repeating or different
// block part from the given position (left to right)
//
// block -> The block
// beg -> begin position
// same -> if true, the first repeating part is searched, otherwise, the first not matching
// return: the searchrecord
function searchRLEDuplicates(block :TJFLBlockContent;
                         beg : byte;
                         same : boolean) : TJFRLEsearch;
var
   i: byte;
   search: TJFRLEsearch;
begin
  for i := beg to length(block)-1 do
  begin
    if ((same = true) AND (block[i] = block[i+1])) then
    begin
      search.repeating := block[i];
      search.position := i;
      search.number := i;
      result := search;
      exit;
    end;
    if ((same = false) AND (block[i] <> block[i+1])) then
    begin
      search.repeating := block[i];
      search.position := i;
      search.number := i;
      result := search;
      exit;
    end;
  end;
  search.repeating := 0;
  search.position := 0;
  search.number:= 0;
  result := search;
end;


// Get Only a part of an Array of chars
function getArrayPart(chars : array of Char) : TOpenCharArray;
var
   i : integer;
   concated : array of Char;
begin
   SetLength(concated, length(chars));
   for i := 0 to length(chars)-1 do
   begin
     concated[i] := chars[i];
   end;
   result := concated;
end;

// Concat two Array of chars
function concatArray(chars1 : array of Char ;
                     chars2 : array of Char
                            ) : TOpenCharArray;
var
   i : integer;
   concated : array of Char;
begin
   SetLength(concated, length(chars1)+length(chars2));
   for i := 0 to length(chars1)-1 do
   begin
     concated[i] := chars1[i];
   end;
   for i := 0 to length(chars2)-1 do
   begin
     concated[length(chars1)+i] := chars2[i];
   end;
   result := concated;
end;

// Move block array content to left, the remaing part is filled with '0' content
procedure shiftArrayContentToLeft(bytes : array of byte; pos : byte; num : byte);
var
   i: byte;
begin
  for i:= pos to num do
  begin
    bytes[i] := bytes[i+num];
  end;
  for i:= num to length(bytes) do
  begin
    bytes[i] := 0;
  end;
end;

function JFLcompress(var block : TJFLBlockContent;var olen : byte) : TJFLBlockContent;
var
    pos,c,t : byte;
    segment : array [1..128] of TJFLSegmentContent;
begin

  // build table
  pos := 0; c := 0;
  while (pos<olen) do
  begin
    if (block[pos] = block[pos+1]) and (pos < olen) then // repeating block
    begin
      c := c + 1;
      while (block[pos] = block[pos+1]) AND (pos < olen) do
      begin
        segment[c].segmenttype := 1;
        segment[c].segmentlen := segment[c].segmentlen + 1; // number of repeating bytes
        segment[c].segment[1] := block[pos];
        pos := pos + 1;
      end;
    end
    else if (block[pos] <> block[pos+1]) and (pos < olen) then  // unique block
    begin
      //pos := pos-1; // In order to process the byte separately
      c := c + 1;
      while (block[pos] <> block[pos+1]) and (pos < olen) do
      begin
        segment[c].segmenttype := 2;
        segment[c].segmentlen := segment[c].segmentlen + 1; // number of repeating bytes
        segment[c].segment[1] := block[pos];
        pos := pos + 1;
      end;
    end;
  end;

  olen := 0;
  for pos := 1 to c do
  begin
    if segment[pos].segmenttype = 1 then    // RLE compress
    begin
      for t:= 0 to segment[pos].segmentlen do
        block[c] := segment[pos].segment[t];
      c := c + segment[pos].segmentlen;
      olen := olen + segment[pos].segmentlen;
    end;
    if segment[pos].segmenttype = 2 then // not compressed
    begin
      for t:= 0 to segment[pos].segmentlen do
        block[c] := segment[pos].segment[t];
      c := c + segment[pos].segmentlen;
      olen := olen + segment[pos].segmentlen;
    end;
  end;

  ShowMessage('Block length='+IntToStr(olen));

end;

// List all content of a directory
// if Dirs is false, then everything but directories are returned and vice versa
// Pattern for searching must be specified also
// Also, either an absolute or relative result is returned
function getDirContent (dirpath :string ;
                        pattern :string;
                        dirs : boolean;
                        absolutemode : boolean) : TStringList;
var
  sl:TStringList;
  SR: TSearchRec;
  i: integer;
begin
  sl := TstringList.Create;

  // To search subdirectories, query must be correct formed
  if dirs then
    if (copy(dirpath,length(dirpath)-1,length(dirpath)) <> '\') then
       dirpath := dirpath + '\';
  if FindFirst(dirpath+pattern, faAnyFile, SR) = 0 then
  begin
    repeat
      if (dirs) then
      begin
        if (SR.Attr = faDirectory) then sl.Add(SR.Name)
      end
      else
      begin
        if (SR.Attr <> faDirectory) then sl.Add(SR.Name);
      end;

    until FindNext(SR) <> 0;
    FindClose(SR);
  end;

  // Delete:
  // - main directory
  // - current directory
  // - directory above
  if (dirs) then
  for i := sl.Count-2 downto 0 do
  begin
    if sl.Strings[i] = '' then sl.Delete(i);
    if sl.Strings[i] = '.' then sl.Delete(i);
    if sl.Strings[i] = '..' then sl.Delete(i);
  end;

  // Append the original absolute path in front
  if (absolutemode=true) then
    for i := 0 to sl.Count-1 do sl.Strings[i] := dirpath + '\' + sl.Strings[i];
  
  //ShowMessage('File0:'+sl.Strings[sl.Count-1]);
  
  result := sl;
end;

// A JFL file is initied for reading
procedure InitJFLForReading(fileName : string);
var i : integer;
begin
     JFLfileStream := TFileStream.Create(fileName, fmOpenRead);
     JFLfileStream.Read(TotalFiles,4);  // Get the record of total files in archive (first 4 bytes)
     //Dec(TotalFiles);                // 0 based, so decript by one
     SetLength(JFLFileEntries,TotalFiles);  // Set the length of the fileentries
     for i:=1 to TotalFiles do JFLfileStream.Read(
              JFLFileEntries[i-1],                // Get the next filename index entry
              SizeOf(JFLFileEntries[i-1]));              // Size of the entries
     JFLFileName := fileName;
end;

// A JFL file is inited for writing as new file. Any existing content is erased.
procedure InitJFLForWriting(fileName : string);
var i : integer;
begin
     JFLfileStream := TFileStream.Create(fileName, fmCreate);
     TotalFiles := 0;
     SetLength(JFLFileEntries,TotalFiles);
     JFLFileName := fileName;
end;

// Release JFL file
procedure UninitJFL();
begin
 JFLfileStream.Free;
end;

// Get the number of file entries inside
function getNumberOfEntries : LongInt;
begin
     result := TotalFiles;
end;

// Get the names of the files inside
function getJFLFileNames : TstringList;
var
   i : LongInt;
   fileNames : TStringList;
   filename : string;
begin
     fileNames := TStringList.Create;
     for i:= 0 to getNumberOfEntries()-1 do
     begin
       fileNames.Add(convertJFLFileNameToString(JFLFileEntries[i].Name));
     end;
     result := fileNames;
end;

// Get the index of the file inside
function getJFLIndexByName (filename : string) : longInt; // returns the position index
var i: longInt;
begin
  for i := 0 to getNumberOfEntries do
  begin
       if (convertJFLFileNameToString(JFLFileEntries[i].Name) = filename) then
       begin
         result := i;
         exit;
       end;
  end;
  result := -1;
end;

// Suggests unique file names for the files, which have the same name in JFL
function unifyFileNames() : TStringList;
var
   s : TStringList;
   i : longint;
   c : longint;
   p : integer;
begin
  s := TstringList.Create;

  // build all the names
  for i:= 0 to getNumberOfEntries-1 do
  begin
    s.Add(convertJFLFileNameToString(JFlFileEntries[i].Name));
  end;
  
  // replace non unique with names with a non-existing name
  c := 0;
  for i:= 0 to getNumberOfEntries-1 do
  begin
    if (s.Find(convertJFLFileNameToString(JFlFileEntries[i].Name),p) = true) and (p <> i)
    then
    begin
      while s.Find(IntToStr(c)+ '_'+convertJFLFileNameToString(JFlFileEntries[i].Name),p) = true do
        c := c + 1;
      s.Strings[i]:= IntToStr(c)+ '_'+convertJFLFileNameToString(JFlFileEntries[i].Name);
    end;
  end;

  result := s;
end;

// Create a descriptor of the file header
function createJFLDescriptionInText (filename: string) : boolean;
var i: longInt;
    s : string;
    outfile : textfile;
    sl : TStringList;
    
begin

 AssignFile(outfile, filename);
 Rewrite(outfile);

 // header - begin
 WriteLn(outFile, '# This line is a comment. This file is a header description file of the JFL file. #');
 WriteLn(outFile,'# This file can be used for ordering of the filenames. #');
 WriteLn(outFile,'# Do not edit manually, unless you know, what you are doing ! #');
 // header - end

  sl := unifyFileNames();

  for i := 0 to getNumberOfEntries-1 do
  begin
    WriteLn(outFile,'Original filename:'+convertJFLFileNameToString(JFlFileEntries[i].Name));
    WriteLn(outFile,'Suggested filename:'+sl.Strings[i]);
    WriteLn(outFile,'Offset:'+IntToStr(JFlFileEntries[i].Offset));
    WriteLn(outFile,'Size:'+IntToStr(JFlFileEntries[i].Size));
    WriteLn(outFile,'Hlam0:'+IntToStr(JFlFileEntries[i].Hlam0));
    WriteLn(outFile,'Hlam1:'+IntToStr(JFlFileEntries[i].Hlam1));
  end;

 CloseFile(outFile);
 
  result := true;
end;

// Generates a description, according to settings
procedure descriptiongenerator();
begin
  if frm_options.chk_desc_create.checked then
    createJFLDescriptionInText(frm_options.txt_descriptor.Text);
end;

// Extract the certain file content inside
function extractJFLContentOfFile (fileName : string; targetDir : string) : string;
var i : longInt;
    outFileStream : TFileStream;
    B,BT,BL,Z: Byte;
    Buffer: array[0..255]of byte;
begin
  i := getJFLIndexByName (fileName);
  if i >= 0 then
  begin
      // logic to read the file content stream block
      outFileStream := TFileStream.Create(targetDir+'\'+fileName, fmCreate or fmShareExclusive);
      memo.Lines.Add(' >>>> '+targetDir+'\'+fileName+' <<<'); // DEBUG
      memo.Lines.Add('Offset:'+
                     IntToStr(JFLFileEntries[i].Offset)+', size:'+
                     IntToStr(JFLFileEntries[i].Size)); // DEBUG
      JFLfileStream.Seek(JFLFileEntries[i].Offset,soFromBeginning);
      memo.Lines.Add('Begin position:'+IntToStr(JFLfileStream.Position)); // DEBUG
      repeat
         JFLfileStream.Read(B,1);
         BT:=B shr 7;   // Compression type (uncompressed or compressed)
         BL:=B and $7F; // How many bytes
         case BT of
           1:
           begin
            JFLfileStream.Read(Z,1);
            FillChar(Buffer,BL+3,Z);
            outFileStream.Write(Buffer,BL+3);
           end;
           0:
           begin
             JFLfileStream.Read(Buffer,BL);
             outFileStream.Write(Buffer,BL);
           end;
         end
      until JFLfileStream.Position>=
            JFLFileEntries[i].Offset+
            JFLFileEntries[i].Size;
      memo.Lines.Add('End position:'+IntToStr(JFLfileStream.Position)); // DEBUG
      memo.Lines.Add('Final file size:'+IntToStr(outfileStream.Position)); // DEBUG
      outFileStream.Free;
  end;
end;

// Creates a full JFL from a directory
procedure createFullJFLFromDirectory(dirpath: string);
var filenames        : TStringList; // stringlist of filenames inside the dir
    t                : longInt; // loops and number of files
    inpfile : TFileStream;          // Temporary content file and input file
    bheader          : byte;        // block header

    repeatc          : byte;        // Repeating character
    repeatcnr        : byte;        // Number of repeating char occurences (128 max)

    buffer128        : TJFLBlockContent;      // The uncompressed block content (max. 128)
    buffer128l       : integer;               // the buffer actual length (128 max)
    b,prevb          : Byte;                  // Two bytes to compare
    l                : Byte;                  // The compressed block length
    
    st               : string;                // temporary string
    
begin

     // Prepare and write simultaneously the index structures
     // and calculate also offsets into temporary file
     // Write the data by offsets, byte by byte

       filenames := getDirContent (dirpath,'\*.*',false,true);
       SetLength(JFlFileEntries,filenames.Count);

       JFLfileStream.Seek(0,soFromBeginning);        // Make dummy header
       JFLfileStream.Write(filenames.count,4);
       for t := 0 to 4+(filenames.Count*SizeOf(JFlFileEntries[t])) do
       begin
          JFLfileStream.WriteByte(0);
       end;
       
       memo.Lines.add('Position:'+IntToStr(JFLfilestream.Size)+
                   ',Filenames count:'+IntToStr(filenames.Count));

       progress.Min:= 0; // DEBUG
       progress.Max:= filenames.count-1;// DEBUG
       for t := 0 to filenames.count-1 do
       begin
              memo.Lines.Add(' <<<< '+filenames.Strings[t]+' >>>>'); // DEBUG
              progress.Position:= t; // DEBUG
              inpfile := TFileStream.Create(filenames.Strings[t],fmOpenRead);
              inpfile.Seek(0,soFromBeginning);
              memo.Lines.Add(' Original Length:'+IntToStr(inpfile.Size)); // DEBUG

              // index building - begin
              JFlFileEntries[t].Name:= ExtractFileName(filenames.Strings[t]); // Set the name of the file
              JFlFileEntries[t].Offset:= JFlFileStream.Position; // Set the offset of the file
              // index building - end

              // File content stream creation - begin
              repeatcnr := 0; buffer128l := 0;
              progress2.Min:= 0; progress2.Max:= inpfile.Size; progress2.Position:= 0;
              while (inpfile.Position < inpfile.Size) do
              begin
                 // 127 decimal equals F7 hexadecimal, it is the highest
                 // number representable by seven bits (01111111)
                 // The bit 0 is a special marking (compressed content or not)
                 if (inpfile.Size-inpfile.Position) < 127 then
                 begin
                   b := inpfile.Size-inpfile.Position;
                   inpfile.Read(buffer128,b);
                   l := b;
                   if frm_options.chk_RLE_compression.Checked then
                     if frm_options.chk_compress_BM.Checked = false then
                       JFLCompress(buffer128,l)// RLE compression - !ALPHA!
                     else
                     begin
                       st := convertJFLFileNameToString(JFlFileEntries[t].Name);
                       if (copy(st,length(st)-3,3) <> '.txt') or
                       (copy(st,length(st)-3,3) <> '.ifc') then
                       JFLCompress(buffer128,l);
                     end;
                   JFLfileStream.Write(l,1);
                   JFLfileStream.Write(buffer128,l);
                 end
                 else
                 begin
                   inpfile.Read(buffer128,127);
                   l := 127;
                   if frm_options.chk_RLE_compression.Checked then
                     if frm_options.chk_compress_BM.Checked = false then
                       JFLCompress(buffer128,l)// RLE compression - !ALPHA!
                     else
                     begin
                       st := convertJFLFileNameToString(JFlFileEntries[t].Name);
                       if (copy(st,length(st)-3,3) <> '.txt') or
                       (copy(st,length(st)-3,3) <> '.ifc') then
                       JFLCompress(buffer128,l);
                     end;
                   JFLfileStream.Write(l,1);
                   JFLfileStream.Write(buffer128,l);
                 end;
              end;
              
              JFlFileEntries[t].Size:= JFlFileStream.Position-JFlFileEntries[t].Offset-1;

              // DEBUG - start
              memo.Lines.Add('Name:'+JFlFileEntries[t].Name); // DEBUG
              memo.Lines.Add('Offset:'+IntToStr(JFlFileEntries[t].Offset)); // DEBUG
              memo.Lines.Add('Size:'+IntToStr(JFlFileEntries[t].Size)); // DEBUG
              // DEBUG - end

       inpfile.Free;

       end;

       // Update the header (offsets and other data)
       JFLfileStream.Seek(4,soFromBeginning);
       for t :=0 to Length(JFlFileEntries)-1 do
         JFLfileStream.Write(
              JFlFileEntries[t],
              SizeOf(JFlFileEntries[t]));
              
end;

// Extracts the full JFL to given directory
procedure extractFullJFLtoDirectory(dirpath: string);
var i : longInt;
    fulldir : string;
begin
  fulldir := dirpath+'\'+ ExtractFileName(JFLFileName);

  if (Pos('.',fulldir)-1) > 0 then
    fulldir:=Copy(fulldir,1,Pos('.',fulldir)-1);   // trimming the extension (1 to 3 before last . sign)

  if not DirectoryExists(fulldir) then
     MkDir(fulldir);
  memo.Lines.add(' Files info : ('+IntToStr(TotalFiles)+') :'); // DEBUG
  progress.Min:= 0; progress.Max:= Totalfiles; progress.Position:= 0; // DEBUG
  descriptiongenerator();
  for i := 0 to Totalfiles-1 do
  begin
    memo.Lines.add(IntToStr(i)+':'+
      convertJFLFileNameToString(JFlFileEntries[i].Name)+','+
      IntToStr(JflFileEntries[i].size)+','+
      IntToStr(JflFileEntries[i].offset)); //DEBUG
    extractJFLContentOfFile (convertJFLFileNameToString(JFlFileEntries[i].Name),fulldir);
    progress.Position:= progress.Position+1; // DEBUG
  end;
  progress.Position:= 0; // DEBUG
end;

// Check is done for following (in header):
// 1. Filenames which occur more than once
// 2. Invalid offsets
//
// Results:
//         0 - all OK
//         1 - invalid offset
//         2 - duplicate filenames
function checkJFLstructure() : TStringList;
var
   i : integer;
   res : TStringList;
   s : string;
begin
   // Invalid offset checks (offsets outside of file)
   res := TStringList.Create();
   for i := 0 to Totalfiles - 1 do
   begin
     if (JFlFileEntries[i].Offset>JFLfileStream.Size) then
     begin
          s := '!1 '+
             IntToStr(JFlFileEntries[i].Offset)+'>'+
             IntToStr(JFlFileEntries[i].Size)+' '+
             convertJFLFileNameToString(JFlFileEntries[i].Name);
          res.Add(s);
     end;
   end;
   if res.Count<1 then res.Add('+0');
   result := res;
end;

end.
